;16-Bit to 7-Segment-Display converter
;display modes:
;  0) unsigned
;  1) signed
;  2) hexadecimal
;  3) inverted, unsigned
;  4) inverted, signed
;  5) inverted, hexadecimal
;ATMega8 @ 1 MHz internal R/C (lfuse=0x81, hfuse=0xD1)
;
;PORTS:
;  PB0-7 = input bits 8~15 (high byte, with internal pullups)
;  PC0   = display 2 (1000's); shift register data
;  PC1   = display 3 (100's); shift register clock
;  PC2   = display 4 (10's)
;  PC3   = display 5 (1's)
;  PC4   = shift register output enable
;  PC5   = toggle display mode (push-button to ground)
;  PD0-7 = input bits 0~7 (low byte, with internal pullups)
;connect display 1 (10000's) cathode to ground through a diode (1N4001)
;shift register: Q0 = a ... Q6 = g, Q7 = minus sign
;
;VERSION HISTORY:
;  2010-07-03  Arne Rossius
;    * BUGFIX: swapped high and low input bytes

.include "../include/m8def.inc"

.def	zero_reg = R13
.def	oldL = R14
.def	oldH = R15

.def	temp = R16
.def	temp2 = R17
.def	inp = R18
.def	disp1 = R19
.def	disp2 = R20
.def	disp3 = R21
.def	disp4 = R22
.def	disp5 = R23
.def	button = R24
.def	mode = R25
.def	timeout = R26
.def	debounce = R27
.def	dataL = R28
.def	dataH = R29

.equ	SEG_A = 1
.equ	SEG_B = 2
.equ	SEG_C = 4
.equ	SEG_D = 8
.equ	SEG_E = 16
.equ	SEG_F = 32
.equ	SEG_G = 64

.equ	FIRST_MINUS = 128
.equ	MINUS  = SEG_G
.equ	ZERO   = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F
.equ	ONE    = SEG_B | SEG_C
.equ	TWO    = SEG_A | SEG_B | SEG_D | SEG_E | SEG_G
.equ	THREE  = SEG_A | SEG_B | SEG_C | SEG_D | SEG_G
.equ	FOUR   = SEG_B | SEG_C | SEG_F | SEG_G
.equ	FIVE   = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G
.equ	SIX    = SEG_A | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	SEVEN  = SEG_A | SEG_B | SEG_C
.equ	EIGHT  = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	NINE   = SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G
.equ	HEX_A  = SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G
.equ	HEX_B  = SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	HEX_C  = SEG_A | SEG_D | SEG_E | SEG_F
.equ	HEX_D  = SEG_B | SEG_C | SEG_D | SEG_E | SEG_G
.equ	HEX_E  = SEG_A | SEG_D | SEG_E | SEG_F | SEG_G
.equ	HEX_F  = SEG_A | SEG_E | SEG_F | SEG_G
.equ	HEX_h  = SEG_C | SEG_E | SEG_F | SEG_G
.equ	MODE_U = SEG_B | SEG_C | SEG_D | SEG_E | SEG_F
.equ	MODE_S = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G
.equ	MODE_H = SEG_B | SEG_C | SEG_E | SEG_F | SEG_G
.equ	MODE_I = SEG_E | SEG_F

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
charset:
	.db ZERO, ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN
	.db EIGHT, NINE, HEX_A, HEX_B, HEX_C, HEX_D, HEX_E, HEX_F

;===============================================================================

reset:
	;stackpointer
	ldi	temp, HIGH(RAMEND)
	out	SPH, temp
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;ports
	ldi	temp, 0x00
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x1F
	out	DDRC, temp
	ldi	temp, 0x20
	out	PORTC, temp
	ldi	temp, 0x00
	out	DDRD, temp
	ldi	temp, 0xFF
	out	PORTD, temp
	
	;registers
	clr	zero_reg
	ldi	button, 0
	ldi	debounce, 0
	
	;load display mode from EEPROM
	ldi	temp, 42
	out	EEARH, zero_reg
	out	EEARL, temp
	sbi	EECR, EERE
	in	mode, EEDR
	cpi	mode, 6
	brlo	PC+3
	ldi	mode, 0
	rcall	mode_write
	
	;display current mode setting
	ldi	disp5, 0 ;blank
	cpi	mode, 3
	brlo	PC+2
	ldi	disp5, MODE_I ;I = Inverted
	
	ldi	disp1, MODE_U ;U = Unsigned
	cpi	mode, 1
	brne	PC+2
	ldi	disp1, MODE_S ;S = Signed
	cpi	mode, 4
	brne	PC+2
	ldi	disp1, MODE_S ;S = Signed
	cpi	mode, 2
	brne	PC+2
	ldi	disp1, MODE_H ;H = Hexadecimal
	cpi	mode, 5
	brne	PC+2
	ldi	disp1, MODE_H ;H = Hexadecimal
	
	ldi	disp2, 0 ;blank
	ldi	disp3, 0 ;blank
	ldi	disp4, 0 ;blank
	ldi	timeout, 170 ;1s timeout
	
;-------------------------------------------------------------------------------
	
loop:
	cpi	timeout, 0
	breq	calc_check
	dec	timeout
	breq	calc
	rjmp	calc_end
calc_check:
	in	dataH, PINB ;High Byte
	in	dataL, PIND ;Low Byte
	cp	oldL, dataL
	cpc	oldH, dataH
	brne	calc
	rjmp	calc_end
calc:
	movw	oldH:oldL, dataH:dataL
	ldi	disp1, 0
	ldi	disp2, 0
	ldi	disp3, 0
	ldi	disp4, 0
	cpi	mode, 3
	brlo	PC+3
	com	dataL ;invert
	com	dataH
	cpi	mode, 1
	breq	signed
	cpi	mode, 4
	breq	signed
	cpi	mode, 2
	breq	hex
	cpi	mode, 5
	breq	hex
	
	;unsigned: clear T bit (no minus sign)
unsigned:
	clt
	rjmp	signed_end
	
signed:
	;signed
	ldi	temp, HIGH(32768) ;positive if < 32768
	cpi	dataL, LOW(32768)
	cpc	dataH, temp
	brlo	unsigned
	;negative
	set
	com	dataL
	com	dataH
	adiw	dataH:dataL, 1 ;2's complement = 1's complement and add 1
signed_end:
	
div10000:
	;10000's
	ldi	temp, HIGH(10000)
div10000_loop:
	cpi	dataL, LOW(10000)
	cpc	dataH, temp
	brlo	div10000_end
	subi	dataL, LOW(10000)
	sbci	dataH, HIGH(10000)
	inc	disp1
	rjmp	div10000_loop
div10000_end:

div1000:
	;1000's
	ldi	temp, HIGH(1000)
div1000_loop:
	cpi	dataL, LOW(1000)
	cpc	dataH, temp
	brlo	div1000_end
	subi	dataL, LOW(1000)
	sbci	dataH, HIGH(1000)
	inc	disp2
	rjmp	div1000_loop
div1000_end:

div100:
	;100's
	cpi	dataL, 100
	cpc	dataH, zero_reg
	brlo	div100_end
	subi	dataL, LOW(100)
	sbci	dataH, HIGH(0)
	inc	disp3
	rjmp	div100
div100_end:

div10:
	;10's
	cpi	dataL, 10
	brlo	div10_end
	subi	dataL, 10
	inc	disp4
	rjmp	div10
div10_end:
	
	;1's
	mov	disp5, dataL
	rjmp	encode
	
hex:
	;hexadecimal
	clt
	ldi	disp1, HEX_h ;display 'h' in front of hex number
	mov	disp2, dataH
	swap	disp2
	andi	disp2, 0x0F
	mov	disp3, dataH
	andi	disp3, 0x0F
	mov	disp4, dataL
	swap	disp4
	andi	disp4, 0x0F
	mov	disp5, dataL
	andi	disp5, 0x0F

	;encode to 7-segment
encode:
	cpi	disp1, HEX_h
	breq	encode_1_end
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp1
	adc	ZH, zero_reg
	lpm	disp1, Z
encode_1_end:
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp2
	adc	ZH, zero_reg
	lpm	disp2, Z
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp3
	adc	ZH, zero_reg
	lpm	disp3, Z
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp4
	adc	ZH, zero_reg
	lpm	disp4, Z
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp5
	adc	ZH, zero_reg
	lpm	disp5, Z
	;zero blanking
	cpi	disp1, ZERO
	brne	zero_end
	ldi	disp1, 0
	cpi	disp2, ZERO
	brne	zero_end
	ldi	disp2, 0
	cpi	disp3, ZERO
	brne	zero_end
	ldi	disp3, 0
	cpi	disp4, ZERO
	brne	zero_end
	ldi	disp4, 0
zero_end:
	;insert minus
	brtc	calc_end
	cpi	disp4, 0
	brne	PC+3
	ldi	disp4, MINUS
	rjmp	calc_end
	cpi	disp3, 0
	brne	PC+3
	ldi	disp3, MINUS
	rjmp	calc_end
	cpi	disp2, 0
	brne	PC+3
	ldi	disp2, MINUS
	rjmp	calc_end
	cpi	disp1, 0
	brne	PC+3
	ldi	disp1, MINUS
	rjmp	calc_end
	ori	disp1, FIRST_MINUS
calc_end:

	;push-button
	in	temp, PINC
	andi	temp, 0x20
	cp	temp, button
	breq	button_debounce
	mov	button, temp
	cpi	debounce, 0
	ldi	debounce, 17 ;100 ms
	brne	button_debounce
	sbrc	temp, 5
	rjmp	button_debounce
	;increment mode
	inc	mode
	cpi	mode, 6
	brlo	PC+2
	ldi	mode, 0
	rcall	mode_write
	rjmp	reset
button_debounce:
	cpi	debounce, 0
	breq	button_end
	dec	debounce
button_end:
	
	;display 1
	mov	temp, disp1
	rcall	shift
	cbi	PORTC, 1
	sbi	PORTC, 4
	rcall	delay
	cbi	PORTC, 4
	
	;display 2
	mov	temp, disp2
	rcall	shift
	cbi	PORTC, 1
	sbi	PORTC, 0
	sbi	PORTC, 4
	rcall	delay
	cbi	PORTC, 4
	
	;display 3
	mov	temp, disp3
	rcall	shift
	;PC1 is HIGH after shifting
	sbi	PORTC, 4
	rcall	delay
	cbi	PORTC, 4
	
	;display 4
	mov	temp, disp4
	rcall	shift
	cbi	PORTC, 1
	sbi	PORTC, 2
	sbi	PORTC, 4
	rcall	delay
	cbi	PORTC, 4
	cbi	PORTC, 2
	
	;display 5
	mov	temp, disp5
	rcall	shift
	cbi	PORTC, 1
	sbi	PORTC, 3
	sbi	PORTC, 4
	rcall	delay
	cbi	PORTC, 4
	cbi	PORTC, 3
	
	rjmp	loop
	
;===============================================================================

shift:
	;output 1 byte to shift register
	ldi	temp2, 8
shift_loop:
	cbi	PORTC, 0
	sbrc	temp, 7
	sbi	PORTC, 0
	cbi	PORTC, 1
	sbi	PORTC, 1
	lsl	temp
	dec	temp2
	brne	shift_loop
	cbi	PORTC, 0
	ret

;--------------------

delay:
	;delay ~1ms
	ldi	temp, 0
delay_loop:
	nop
	dec	temp
	brne	delay_loop
	ret
	
;--------------------

mode_write:
	;write current mode to eeprom
	ldi	temp, 42
	out	EEARH, zero_reg
	out	EEARL, temp
	out	EEDR, mode
	sbi	EECR, EEMWE
	sbi	EECR, EEWE
mode_write_wait:
	sbic	EECR, EEWE
	rjmp	mode_write_wait
	ret
	